@@ -1,15 +1,15 @@
 import logging
-import re
+import shlex
 import subprocess
 import sys
 import tempfile
 import time
 
 from contextlib import contextmanager
-from dataclasses import dataclass
+from dataclasses import dataclass, fields, is_dataclass
 from enum import IntEnum
 from pathlib import Path
-from typing import List, Optional, Sequence, Tuple, Union
+from typing import List, Optional, Sequence, TextIO, Union
 
 from remote.exceptions import InvalidInputError
 
@@ -17,6 +17,8 @@ from .exceptions import RemoteConnectionError, RemoteExecutionError
 
 logger = logging.getLogger(__name__)
 
+DEFAULT_SSH_PORT = 22
+
 
 def _temp_file(lines: List[str]) -> Path:
     """Create a temporary file with provided content and return its path
@@ -54,7 +56,28 @@ class ForwardingOptions:
 
     remote_port: int
     local_port: int
-    host: str = "localhost"
+    remote_interface: str = "localhost"
+    local_interface: Optional[str] = None
+
+    @classmethod
+    def from_string(cls, port_args: str) -> "ForwardingOptions":
+        """Parse port values from the user input.
+        :param host: the input string from port tunnelling option.
+        :returns: A tuple of remote port, local port.
+        """
+        ports: List = port_args.split(":")
+        if len(ports) > 2:
+            raise InvalidInputError("Please pass a valid value to enable local port forwarding")
+        try:
+            if len(ports) == 1:
+                return cls(int(ports[0]), int(ports[0]))
+            return cls(int(ports[0]), int(ports[1]))
+        except ValueError as e:
+            raise InvalidInputError("Please pass valid integer value for ports") from e
+
+    def to_ssh_string(self) -> str:
+        prefix = f"{self.local_interface}:" if self.local_interface else ""
+        return f"{prefix}{self.local_port}:{self.remote_interface}:{self.remote_port}"
 
 
 class VerbosityLevel(IntEnum):
@@ -63,16 +86,25 @@ class VerbosityLevel(IntEnum):
     VERBOSE = 3
 
 
+@dataclass(frozen=True)
+class CommunicationOptions:
+    stdin: Optional[TextIO] = sys.stdin
+    stdout: TextIO = sys.stdout
+    stderr: TextIO = sys.stderr
+
+
 @dataclass(frozen=True)
 class Ssh:
     """Ssh configuration class, pregenrates and executes commands remotely"""
 
     host: str
+    port: Optional[int] = None
     force_tty: bool = True
     verbosity_level: VerbosityLevel = VerbosityLevel.QUIET
     use_gssapi_auth: bool = True
     disable_password_auth: bool = True
     local_port_forwarding: Optional[ForwardingOptions] = None
+    communication: CommunicationOptions = CommunicationOptions()
 
     def generate_command(self) -> List[str]:
         """Generate the base ssh command to execute (without host)"""
@@ -89,17 +121,11 @@ class Ssh:
             command.append(f"-{options}")
         if self.disable_password_auth:
             command.extend(("-o", "BatchMode=yes"))
+        if self.port and self.port != DEFAULT_SSH_PORT:
+            command.extend(("-p", str(self.port)))
 
         if self.local_port_forwarding:
-            command.extend(
-                (
-                    "-L",
-                    (
-                        f"{self.local_port_forwarding.local_port}:"
-                        f"{self.local_port_forwarding.host}:{self.local_port_forwarding.remote_port}"
-                    ),
-                )
-            )
+            command.extend(("-L", self.local_port_forwarding.to_ssh_string()))
 
         return command
 
@@ -107,11 +133,10 @@ class Ssh:
         """Generate the base ssh command to execute (without host)"""
         return prepare_shell_command(self.generate_command())
 
-    def execute(self, command: str, dry_run: bool = False, raise_on_error: bool = True) -> int:
+    def execute(self, command: str, raise_on_error: bool = True) -> int:
         """Execute a command remotely using SSH and return it's exit code
 
         :param command: a command to execute
-        :param dry_run: log command instead of executing it
         :param raise_on_error: raise an exception is remote execution
 
         :returns: exit code of remote command or 255 if connection didn't go through
@@ -119,12 +144,14 @@ class Ssh:
         subprocess_command = self.generate_command()
 
         logger.info("Executing:\n%s %s <<EOS\n%sEOS", " ".join(subprocess_command), self.host, command)
-        if dry_run:
-            return 0
-
         subprocess_command.extend((self.host, command))
         with _measure_duration("Execution"):
-            result = subprocess.run(subprocess_command, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin)
+            result = subprocess.run(
+                subprocess_command,
+                stdout=self.communication.stdout,
+                stderr=self.communication.stderr,
+                stdin=self.communication.stdin,
+            )
 
         if raise_on_error:
             # ssh exits with the exit status of the remote command or with 255 if an error occurred
@@ -147,6 +174,7 @@ def rsync(
     excludes: List[str] = None,
     includes: List[str] = None,
     extra_args: List[str] = None,
+    communication=CommunicationOptions(),
 ):
     """Run rsync to sync files from src into dst
 
@@ -162,6 +190,7 @@ def rsync(
     :param excludes: List of file patterns to exclude from syncing
     :param includes: List of file patterns to include even if they were excluded by exclude filters
     :param extra_args: Extra arguments for rsync function
+    :param communication: file descriptors to use for process communication
     """
 
     logger.info("Sync files from %s to %s", src, dst)
@@ -188,7 +217,7 @@ def rsync(
 
     logger.info("Starting sync with command %s", " ".join(args))
     with _measure_duration("Sync"):
-        result = subprocess.run(args, stdout=sys.stdout, stderr=sys.stderr)
+        result = subprocess.run(args, stdout=communication.stdout, stderr=communication.stderr)
 
     for file in cleanup:
         file.unlink()
@@ -205,31 +234,25 @@ def prepare_shell_command(command: Union[str, Sequence[str]]) -> str:
     if len(command) == 1 and " " in command[0]:
         return command[0]
 
+    return " ".join([shlex.quote(c) for c in command])
+
+
+def pformat_dataclass(obj, indent="  "):
+    """Return a string with an object contents prettified"""
     result = []
-    for item in command:
-        if not item:
-            continue
-        if re.search(r"\s+", item):
-            result.append(f"'{item}'")
+
+    has_dataclass_fields = False
+    for field in fields(obj):
+        value = getattr(obj, field.name)
+        if is_dataclass(value):
+            str_value = "\n" + pformat_dataclass(value, indent + "  ")
+            has_dataclass_fields = True
         else:
-            result.append(item)
+            str_value = str(value)
+        result.append((field.name, str_value))
 
-    return " ".join(result)
-
-
-def parse_ports(port_args: Optional[str]) -> Optional[Tuple[int, int]]:
-    """Parse port values from the user input.
-    :param host: the input string from port tunnelling option.
-    :returns: A tuple of remote port, local port.
-    """
-    if not port_args:
-        return None
-    ports: List = port_args.split(":")
-    if len(ports) > 2:
-        raise InvalidInputError("Please pass a valid value to enable local port forwarding")
-    try:
-        if len(ports) == 1:
-            return (int(ports[0]), int(ports[0]))
-        return (int(ports[0]), int(ports[1]))
-    except ValueError as e:
-        raise InvalidInputError("Please pass valid integer value for ports") from e
+    if has_dataclass_fields:
+        return "\n".join(f"{indent}- {name}: {value}" for name, value in result)
+    else:
+        width = max(len(name) for name, _ in result)
+        return "\n".join(f"{indent}- {name: <{width}}: {value}" for name, value in result)
