@@ -4,9 +4,11 @@ We only mock subprocess calls since we cannot do multi-host testing.
 Some of the test above don't verify much, but they at least ensure that all parts work well together.
 """
 import os
+import sys
 
 from contextlib import contextmanager
-from unittest.mock import ANY, Mock, call, patch
+from datetime import datetime
+from unittest.mock import ANY, MagicMock, Mock, call, patch
 
 import pytest
 
@@ -15,6 +17,7 @@ from click.testing import CliRunner
 
 from remote import entrypoints
 from remote.configuration.classic import CONFIG_FILE_NAME, IGNORE_FILE_NAME, INDEX_FILE_NAME
+from remote.configuration.toml import WORKSPACE_CONFIG
 from remote.exceptions import RemoteExecutionError
 
 TEST_HOST = "test-host1.example.com"
@@ -93,6 +96,10 @@ def test_validate_connection_string(connection, is_valid):
 
 
 @patch("remote.util.subprocess.run")
+@patch(
+    "remote.configuration.toml.TomlConfigurationMedium.generate_remote_directory",
+    MagicMock(return_value=".remotes/myproject_foo"),
+)
 def test_remote_init(mock_run, tmp_path):
     mock_run.return_value = Mock(returncode=0)
     subdir = tmp_path / "myproject"
@@ -103,25 +110,37 @@ def test_remote_init(mock_run, tmp_path):
         result = runner.invoke(entrypoints.remote_init, ["test-host.example.com"])
 
     assert result.exit_code == 0
-    assert "Created remote directory at test-host.example.com:.remotes/myproject_" in result.output
+    assert "Created remote directory at test-host.example.com:.remotes/myproject_foo" in result.output
     assert "Remote is configured and ready to use" in result.output
 
     mock_run.assert_called_once_with(
-        ["ssh", "-tKq", "-o", "BatchMode=yes", "test-host.example.com", ANY], stdin=ANY, stdout=ANY, stderr=ANY
+        ["ssh", "-tKq", "-o", "BatchMode=yes", "test-host.example.com", ANY],
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
-    assert (subdir / CONFIG_FILE_NAME).exists()
-    assert (subdir / CONFIG_FILE_NAME).read_text().startswith("test-host.example.com:.remotes/myproject_")
-    assert (subdir / IGNORE_FILE_NAME).exists()
+    assert (subdir / WORKSPACE_CONFIG).exists()
     assert (
-        (subdir / IGNORE_FILE_NAME).read_text()
+        (subdir / WORKSPACE_CONFIG).read_text()
         == """\
-pull:
-push:
-both:
-.remote
-.remoteignore
-.remoteindex
+[[hosts]]
+host = "test-host.example.com"
+directory = ".remotes/myproject_foo"
+default = true
+supports_gssapi_auth = true
+
+[push]
+exclude = []
+include = []
+
+[pull]
+exclude = []
+include = []
+
+[both]
+exclude = [ ".remote.toml",]
+include = []
 """
     )
 
@@ -147,27 +166,67 @@ Remote is configured and ready to use
 
     mock_run.assert_called_once_with(
         ["ssh", "-tKq", "-o", "BatchMode=yes", "test-host.example.com", "mkdir -p .path/test.dir/_test-dir"],
-        stdin=ANY,
-        stdout=ANY,
-        stderr=ANY,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
-    assert (subdir / CONFIG_FILE_NAME).exists()
-    assert (subdir / CONFIG_FILE_NAME).read_text() == "test-host.example.com:.path/test.dir/_test-dir\n"
-    assert (subdir / IGNORE_FILE_NAME).exists()
+    assert (subdir / WORKSPACE_CONFIG).exists()
     assert (
-        (subdir / IGNORE_FILE_NAME).read_text()
+        (subdir / WORKSPACE_CONFIG).read_text()
         == """\
-pull:
-push:
-both:
-.remote
-.remoteignore
-.remoteindex
+[[hosts]]
+host = "test-host.example.com"
+directory = ".path/test.dir/_test-dir"
+default = true
+supports_gssapi_auth = true
+
+[push]
+exclude = []
+include = []
+
+[pull]
+exclude = []
+include = []
+
+[both]
+exclude = [ ".remote.toml",]
+include = []
 """
     )
 
 
+@patch("remote.util.subprocess.run")
+def test_remote_init_gitignore(mock_run, tmp_path):
+    mock_run.return_value = Mock(returncode=0)
+    subdir = tmp_path / "myproject"
+    subdir.mkdir()
+    (subdir / ".git").mkdir()
+
+    runner = CliRunner()
+    with cwd(subdir):
+        result = runner.invoke(entrypoints.remote_init, ["test-host.example.com:.path/test.dir/_test-dir/"])
+
+    assert result.exit_code == 0
+    assert ".remote*" in (subdir / ".gitignore").read_text()
+
+
+@patch("remote.util.subprocess.run")
+def test_remote_init_gitignore_no_double_writing(mock_run, tmp_path):
+    mock_run.return_value = Mock(returncode=0)
+    subdir = tmp_path / "myproject"
+    subdir.mkdir()
+    (subdir / ".git").mkdir()
+    (subdir / ".gitignore").write_text("some\nbuild\n.remote*\n.gradle\n")
+
+    runner = CliRunner()
+    with cwd(subdir):
+        result = runner.invoke(entrypoints.remote_init, ["test-host.example.com:.path/test.dir/_test-dir/"])
+
+    assert result.exit_code == 0
+    assert "some\nbuild\n.remote*\n.gradle\n" == (subdir / ".gitignore").read_text()
+
+
 @patch("remote.util.subprocess.run")
 def test_remote_init_fails_after_ssh_error(mock_run, tmp_path):
     mock_run.return_value = Mock(returncode=255)
@@ -187,8 +246,7 @@ Please check if host is accessible via SSH
 """
     )
 
-    assert not (subdir / CONFIG_FILE_NAME).exists()
-    assert not (subdir / IGNORE_FILE_NAME).exists()
+    assert not (subdir / WORKSPACE_CONFIG).exists()
 
 
 def test_remote_init_fails_if_workspace_is_already_initated(tmp_workspace):
@@ -200,7 +258,7 @@ def test_remote_init_fails_if_workspace_is_already_initated(tmp_workspace):
     assert (
         result.output
         == """\
-A configured workspace already exists in the current directory.
+A configured workspace already exists in the current working directory.
 If you want to add a new host to it, please use remote-add.
 """
     )
@@ -216,7 +274,8 @@ def test_remote_init_fails_on_input_validation(tmp_path):
         result = runner.invoke(entrypoints.remote_init, ["host:path:path"])
 
     assert result.exit_code == 2
-    assert not (tmp_path / CONFIG_FILE_NAME).exists()
+
+    assert not (tmp_path / WORKSPACE_CONFIG).exists()
 
 
 def test_remote_commands_fail_on_no_workspace(tmp_path):
@@ -261,7 +320,10 @@ def test_remote_add_adds_host(mock_run, tmp_workspace):
     assert (tmp_workspace / CONFIG_FILE_NAME).read_text() == f"{TEST_CONFIG}\nhost:directory\n"
 
     mock_run.assert_called_once_with(
-        ["ssh", "-tKq", "-o", "BatchMode=yes", "host", "mkdir -p directory"], stdin=ANY, stdout=ANY, stderr=ANY,
+        ["ssh", "-tKq", "-o", "BatchMode=yes", "host", "mkdir -p directory"],
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
 
@@ -375,6 +437,85 @@ def test_remote(mock_run, tmp_workspace):
                     "--delete",
                     "--rsync-path",
                     "mkdir -p .remotes/myproject && rsync",
+                    "--include-from",
+                    ANY,
+                    "--exclude-from",
+                    ANY,
+                    f"{tmp_workspace}/",
+                    f"{TEST_HOST}:{TEST_DIR}",
+                ],
+                stdout=sys.stdout,
+                stderr=sys.stderr,
+            ),
+            call(
+                [
+                    "ssh",
+                    "-tKq",
+                    "-o",
+                    "BatchMode=yes",
+                    TEST_HOST,
+                    """\
+cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
+echo test >> .file
+""",
+                ],
+                stdout=sys.stdout,
+                stdin=sys.stdin,
+                stderr=sys.stderr,
+            ),
+            call(
+                [
+                    "rsync",
+                    "-arlpmchz",
+                    "--copy-unsafe-links",
+                    "-e",
+                    "ssh -Kq -o BatchMode=yes",
+                    "--force",
+                    "--exclude-from",
+                    ANY,
+                    f"{TEST_HOST}:{TEST_DIR}/",
+                    f"{tmp_workspace}",
+                ],
+                stdout=sys.stdout,
+                stderr=sys.stderr,
+            ),
+        ]
+    )
+
+
+@patch(
+    "remote.entrypoints.datetime",
+    MagicMock(now=MagicMock(return_value=datetime(year=2020, month=7, day=13, hour=10, minute=11, second=12))),
+)
+@patch("remote.util.subprocess.run")
+def test_remote_with_output_logging(mock_run, tmp_workspace):
+    mock_run.return_value = Mock(returncode=0)
+    runner = CliRunner()
+
+    with cwd(tmp_workspace):
+        result = runner.invoke(entrypoints.remote, ["--log", "my_logs", "echo test >> .file"])
+
+    assert result.exit_code == 0
+    assert mock_run.call_count == 3
+    mock_run.assert_has_calls(
+        [
+            call(
+                [
+                    "rsync",
+                    "-arlpmchz",
+                    "--copy-unsafe-links",
+                    "-e",
+                    "ssh -Kq -o BatchMode=yes",
+                    "--force",
+                    "--delete",
+                    "--rsync-path",
+                    "mkdir -p .remotes/myproject && rsync",
+                    "--include-from",
+                    ANY,
                     "--exclude-from",
                     ANY,
                     f"{tmp_workspace}/",
@@ -390,15 +531,17 @@ def test_remote(mock_run, tmp_workspace):
                     "-o",
                     "BatchMode=yes",
                     TEST_HOST,
-                    """if [ -f .remotes/myproject/.remoteenv ]; then
-  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
-fi
+                    """\
 cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
 echo test >> .file
 """,
                 ],
                 stdout=ANY,
-                stdin=ANY,
+                stdin=None,
                 stderr=ANY,
             ),
             call(
@@ -419,6 +562,202 @@ echo test >> .file
             ),
         ]
     )
+    for mock_call in mock_run.mock_calls:
+        name, args, kwargs = mock_call
+        assert kwargs["stderr"].name.endswith("my_logs/2020-07-13_10:11:12/test-host1.example.com_output.log")
+        assert kwargs["stdout"].name.endswith("my_logs/2020-07-13_10:11:12/test-host1.example.com_output.log")
+        try:
+            assert kwargs["stdin"] is None
+        except KeyError:
+            pass
+
+
+@patch(
+    "remote.entrypoints.datetime",
+    MagicMock(now=MagicMock(return_value=datetime(year=2020, month=7, day=13, hour=10, minute=11, second=12))),
+)
+@patch("remote.util.subprocess.run")
+def test_remote_mass(mock_run, tmp_workspace):
+    mock_run.return_value = Mock(returncode=0)
+    runner = CliRunner()
+
+    with cwd(tmp_workspace):
+        result = runner.invoke(entrypoints.remote, ["--multi", "echo test >> .file"])
+
+    assert result.exit_code == 0
+    assert mock_run.call_count == 3
+    mock_run.assert_has_calls(
+        [
+            call(
+                [
+                    "rsync",
+                    "-arlpmchz",
+                    "--copy-unsafe-links",
+                    "-e",
+                    "ssh -Kq -o BatchMode=yes",
+                    "--force",
+                    "--delete",
+                    "--rsync-path",
+                    "mkdir -p .remotes/myproject && rsync",
+                    "--include-from",
+                    ANY,
+                    "--exclude-from",
+                    ANY,
+                    f"{tmp_workspace}/",
+                    f"{TEST_HOST}:{TEST_DIR}",
+                ],
+                stdout=ANY,
+                stderr=ANY,
+            ),
+            call(
+                [
+                    "ssh",
+                    "-tKq",
+                    "-o",
+                    "BatchMode=yes",
+                    TEST_HOST,
+                    """\
+cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
+echo test >> .file
+""",
+                ],
+                stdout=ANY,
+                stdin=None,
+                stderr=ANY,
+            ),
+            call(
+                [
+                    "rsync",
+                    "-arlpmchz",
+                    "--copy-unsafe-links",
+                    "-e",
+                    "ssh -Kq -o BatchMode=yes",
+                    "--force",
+                    "--exclude-from",
+                    ANY,
+                    f"{TEST_HOST}:{TEST_DIR}/",
+                    f"{tmp_workspace}",
+                ],
+                stdout=ANY,
+                stderr=ANY,
+            ),
+        ]
+    )
+    for mock_call in mock_run.mock_calls:
+        name, args, kwargs = mock_call
+        assert kwargs["stderr"].name.endswith("logs/2020-07-13_10:11:12/test-host1.example.com_output.log")
+        assert kwargs["stdout"].name.endswith("logs/2020-07-13_10:11:12/test-host1.example.com_output.log")
+        try:
+            assert kwargs["stdin"] is None
+        except KeyError:
+            pass
+
+
+@pytest.mark.parametrize("label, host", [("usual", "host1"), ("unusual", "host2"), ("2", "host2"), ("3", "host3")])
+@patch("remote.util.subprocess.run")
+def test_remote_labeling_works(mock_run, tmp_path, label, host):
+    mock_run.return_value = Mock(returncode=0)
+    runner = CliRunner()
+    (tmp_path / WORKSPACE_CONFIG).write_text(
+        f"""\
+[[hosts]]
+host = "host1"
+directory = "{TEST_DIR}"
+default = true
+label = "usual"
+
+[[hosts]]
+host = "host2"
+directory = "{TEST_DIR}"
+label = "unusual"
+
+[[hosts]]
+host = "host3"
+directory = "{TEST_DIR}"
+"""
+    )
+
+    with cwd(tmp_path):
+        result = runner.invoke(entrypoints.remote, ["-l", label, "echo test >> .file"])
+
+    assert result.exit_code == 0
+    assert mock_run.call_count == 3
+    mock_run.assert_has_calls(
+        [
+            call(
+                [
+                    "rsync",
+                    "-arlpmchz",
+                    "--copy-unsafe-links",
+                    "-e",
+                    "ssh -Kq -o BatchMode=yes",
+                    "--force",
+                    "--delete",
+                    "--rsync-path",
+                    "mkdir -p .remotes/myproject && rsync",
+                    "--include-from",
+                    ANY,
+                    "--exclude-from",
+                    ANY,
+                    f"{tmp_path}/",
+                    f"{host}:{TEST_DIR}",
+                ],
+                stdout=sys.stdout,
+                stderr=sys.stderr,
+            ),
+            call(
+                [
+                    "ssh",
+                    "-tKq",
+                    "-o",
+                    "BatchMode=yes",
+                    host,
+                    """\
+cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
+echo test >> .file
+""",
+                ],
+                stdout=sys.stdout,
+                stdin=sys.stdin,
+                stderr=sys.stderr,
+            ),
+            call(
+                [
+                    "rsync",
+                    "-arlpmchz",
+                    "--copy-unsafe-links",
+                    "-e",
+                    "ssh -Kq -o BatchMode=yes",
+                    "--force",
+                    "--exclude-from",
+                    ANY,
+                    f"{host}:{TEST_DIR}/",
+                    f"{tmp_path}",
+                ],
+                stdout=sys.stdout,
+                stderr=sys.stderr,
+            ),
+        ]
+    )
+
+
+@patch("remote.util.subprocess.run")
+def test_remote_fails_on_unknown_option(mock_run, tmp_workspace):
+    runner = CliRunner()
+
+    with cwd(tmp_workspace):
+        result = runner.invoke(entrypoints.remote, ["--unknown-opt", "echo", "test >> .file"])
+
+    assert result.exit_code == 2
+    assert "Error: no such option --unknown-opt" in result.output
 
 
 @patch("remote.util.subprocess.run")
@@ -444,13 +783,15 @@ def test_remote_execution_fail(mock_run, tmp_workspace):
                     "--delete",
                     "--rsync-path",
                     "mkdir -p .remotes/myproject && rsync",
+                    "--include-from",
+                    ANY,
                     "--exclude-from",
                     ANY,
                     f"{tmp_workspace}/",
                     f"{TEST_HOST}:{TEST_DIR}",
                 ],
-                stdout=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stderr=sys.stderr,
             ),
             call(
                 [
@@ -459,16 +800,18 @@ def test_remote_execution_fail(mock_run, tmp_workspace):
                     "-o",
                     "BatchMode=yes",
                     TEST_HOST,
-                    """if [ -f .remotes/myproject/.remoteenv ]; then
-  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
-fi
+                    """\
 cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
 echo 'test >> .file'
 """,
                 ],
-                stdout=ANY,
-                stdin=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stdin=sys.stdin,
+                stderr=sys.stderr,
             ),
             call(
                 [
@@ -483,8 +826,8 @@ echo 'test >> .file'
                     f"{TEST_HOST}:{TEST_DIR}/",
                     f"{tmp_workspace}",
                 ],
-                stdout=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stderr=sys.stderr,
             ),
         ]
     )
@@ -499,7 +842,7 @@ def test_remote_sync_fail(mock_run, tmp_workspace):
     with cwd(tmp_workspace):
         result = runner.invoke(entrypoints.remote, ["echo test >> .file"])
 
-    assert result.exit_code == 1
+    assert result.exit_code == 255
     mock_run.assert_called_once_with(
         [
             "rsync",
@@ -511,13 +854,15 @@ def test_remote_sync_fail(mock_run, tmp_workspace):
             "--delete",
             "--rsync-path",
             "mkdir -p .remotes/myproject && rsync",
+            "--include-from",
+            ANY,
             "--exclude-from",
             ANY,
             f"{tmp_workspace}/",
             f"{TEST_HOST}:{TEST_DIR}",
         ],
-        stdout=ANY,
-        stderr=ANY,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
 
@@ -537,19 +882,32 @@ def test_remote_quick(mock_run, tmp_workspace):
             "-o",
             "BatchMode=yes",
             TEST_HOST,
-            """if [ -f .remotes/myproject/.remoteenv ]; then
-  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
-fi
+            """\
 cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
 echo test
 """,
         ],
-        stdout=ANY,
-        stdin=ANY,
-        stderr=ANY,
+        stdout=sys.stdout,
+        stdin=sys.stdin,
+        stderr=sys.stderr,
     )
 
 
+@patch("remote.util.subprocess.run")
+def test_remote_quick_fails_on_unknown_option(mock_run, tmp_workspace):
+    runner = CliRunner()
+
+    with cwd(tmp_workspace):
+        result = runner.invoke(entrypoints.remote_quick, ["--unknown-opt", "echo", "test >> .file"])
+
+    assert result.exit_code == 2
+    assert "Error: no such option --unknown-opt" in result.output
+
+
 @patch("remote.util.subprocess.run")
 def test_remote_quick_execution_fail(mock_run, tmp_workspace):
     mock_run.return_value = Mock(returncode=15)
@@ -566,16 +924,18 @@ def test_remote_quick_execution_fail(mock_run, tmp_workspace):
             "-o",
             "BatchMode=yes",
             TEST_HOST,
-            """if [ -f .remotes/myproject/.remoteenv ]; then
-  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
-fi
+            """\
 cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
 echo test
 """,
         ],
-        stdout=ANY,
-        stdin=ANY,
-        stderr=ANY,
+        stdout=sys.stdout,
+        stdin=sys.stdin,
+        stderr=sys.stderr,
     )
 
 
@@ -600,13 +960,15 @@ def test_remote_push(mock_run, tmp_workspace):
             "--delete",
             "--rsync-path",
             "mkdir -p .remotes/myproject && rsync",
+            "--include-from",
+            ANY,
             "--exclude-from",
             ANY,
             f"{tmp_workspace}/",
             f"{TEST_HOST}:{TEST_DIR}",
         ],
-        stdout=ANY,
-        stderr=ANY,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
 
@@ -618,7 +980,7 @@ def test_remote_push_mass(mock_run, tmp_workspace):
     runner = CliRunner()
 
     with cwd(tmp_workspace):
-        result = runner.invoke(entrypoints.remote_push, "--mass")
+        result = runner.invoke(entrypoints.remote_push, "--multi")
 
     assert result.exit_code == 0
     assert mock_run.call_count == 2
@@ -636,13 +998,15 @@ def test_remote_push_mass(mock_run, tmp_workspace):
                     "--delete",
                     "--rsync-path",
                     "mkdir -p .remotes/myproject && rsync",
+                    "--include-from",
+                    ANY,
                     "--exclude-from",
                     ANY,
                     f"{tmp_workspace}/",
                     f"{TEST_HOST}:{TEST_DIR}",
                 ],
-                stdout=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stderr=sys.stderr,
             ),
             call(
                 [
@@ -656,13 +1020,15 @@ def test_remote_push_mass(mock_run, tmp_workspace):
                     "--delete",
                     "--rsync-path",
                     "mkdir -p other-directory && rsync",
+                    "--include-from",
+                    ANY,
                     "--exclude-from",
                     ANY,
                     f"{tmp_workspace}/",
                     "new-host:other-directory",
                 ],
-                stdout=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stderr=sys.stderr,
             ),
         ]
     )
@@ -691,8 +1057,8 @@ def test_remote_pull(mock_run, tmp_workspace):
             f"{TEST_HOST}:{TEST_DIR}/",
             str(tmp_workspace),
         ],
-        stdout=ANY,
-        stderr=ANY,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
 
@@ -720,8 +1086,8 @@ def test_remote_pull_subdirs(mock_run, tmp_workspace):
                     f"{TEST_HOST}:{TEST_DIR}/build",
                     f"{tmp_workspace}/",
                 ],
-                stdout=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stderr=sys.stderr,
             ),
             call(
                 [
@@ -735,8 +1101,8 @@ def test_remote_pull_subdirs(mock_run, tmp_workspace):
                     f"{TEST_HOST}:{TEST_DIR}/dist",
                     f"{tmp_workspace}/",
                 ],
-                stdout=ANY,
-                stderr=ANY,
+                stdout=sys.stdout,
+                stderr=sys.stderr,
             ),
         ]
     )
@@ -752,7 +1118,10 @@ def test_remote_delete(mock_run, tmp_workspace):
 
     assert result.exit_code == 0
     mock_run.assert_called_once_with(
-        ["ssh", "-tKq", "-o", "BatchMode=yes", TEST_HOST, f"rm -rf {TEST_DIR}"], stdin=ANY, stdout=ANY, stderr=ANY,
+        ["ssh", "-tKq", "-o", "BatchMode=yes", TEST_HOST, f"rm -rf {TEST_DIR}"],
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
 
@@ -799,14 +1168,54 @@ def test_remote_port_forwarding_successful(
                 "-L",
                 expected_port_forwarding,
                 "test-host1.example.com",
-                """if [ -f .remotes/myproject/.remoteenv ]; then
-  source .remotes/myproject/.remoteenv 2>/dev/null 1>/dev/null
-fi
+                """\
 cd .remotes/myproject
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd .
 echo test
 """,
             ],
-            stderr=ANY,
-            stdin=ANY,
-            stdout=ANY,
+            stderr=sys.stderr,
+            stdin=sys.stdin,
+            stdout=sys.stdout,
         )
+
+
+@patch("remote.util.subprocess.run")
+def test_stream_changes(mock_run, tmp_workspace):
+    """Ensure the execution with stream changes runs successfully"""
+    mock_run.return_value = Mock(returncode=0)
+    runner = CliRunner()
+    with cwd(tmp_workspace):
+        result = runner.invoke(entrypoints.remote, ["--stream-changes", "echo test"])
+        assert result.exit_code == 0
+
+
+@patch("remote.explain.subprocess.run")
+@patch("remote.util.subprocess.run")
+def test_remote_explain(util_run, explain_run, tmp_workspace):
+    # This is jsut a smoke test to check that remote-explain doesn't throw any exceptions
+    # It is pretty hard to unit-test it correctly
+    util_run.return_value = Mock(returncode=0)
+    explain_run.return_value = Mock(
+        returncode=0,
+        stdout="""\
+PING some-host.example.com (1.1.1.1): 56 data bytes
+64 bytes from 1.1.1.1: icmp_seq=0 ttl=59 time=25.608 ms
+64 bytes from 1.1.1.1: icmp_seq=1 ttl=59 time=15.121 ms
+64 bytes from 1.1.1.1: icmp_seq=2 ttl=59 time=15.735 ms
+
+--- some-host.example.com ping statistics ---
+3 packets transmitted, 3 packets received, 0.0% packet loss
+round-trip min/avg/max/stddev = 15.121/18.821/25.608/4.805 ms
+""",
+    )
+    runner = CliRunner()
+    with cwd(tmp_workspace):
+        result = runner.invoke(entrypoints.remote_explain, ["--deep"])
+
+    explain_run.assert_has_calls([call(["ping", "-c", "10", "test-host1.example.com"], capture_output=True, text=True)])
+    explain_run.assert_has_calls([call(["ping", "-c", "1", "test-host1.example.com"], capture_output=True, text=True)])
+    assert result.exit_code == 0
