@@ -1,11 +1,13 @@
-from unittest.mock import ANY, MagicMock, patch
+import sys
+
+from unittest.mock import MagicMock, patch
 
 import pytest
 
 from pytest import raises
 
 from remote.exceptions import InvalidInputError, RemoteConnectionError, RemoteExecutionError
-from remote.util import ForwardingOptions, Ssh, VerbosityLevel, _temp_file, parse_ports, prepare_shell_command, rsync
+from remote.util import ForwardingOptions, Ssh, VerbosityLevel, _temp_file, prepare_shell_command, rsync
 
 
 def test_temp_file():
@@ -96,8 +98,8 @@ def test_rsync_respects_all_options(mock_run, rsync_ssh):
             "src/",
             "dst",
         ],
-        stdout=ANY,
-        stderr=ANY,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
     )
 
 
@@ -140,6 +142,7 @@ def test_rsync_always_removes_temporary_files(mock_temp_file, mock_run, returnco
     "ssh, expected_cmd",
     [
         (Ssh("host"), "ssh -tKq -o BatchMode=yes"),
+        (Ssh("host", port=12345, force_tty=False), "ssh -Kq -o BatchMode=yes -p 12345"),
         (Ssh("host", force_tty=False), "ssh -Kq -o BatchMode=yes"),
         (Ssh("host", disable_password_auth=False), "ssh -tKq"),
         (Ssh("host", verbosity_level=VerbosityLevel.DEFAULT), "ssh -tK -o BatchMode=yes"),
@@ -195,7 +198,7 @@ def test_ssh_execute(mock_run, ports, expected_command_run):
     code = ssh.execute("exit 0")
 
     assert code == 0
-    mock_run.assert_called_once_with(expected_command_run, stdout=ANY, stderr=ANY, stdin=ANY)
+    mock_run.assert_called_once_with(expected_command_run, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin)
 
 
 @pytest.mark.parametrize("returncode, error", [(255, RemoteConnectionError), (1, RemoteExecutionError)])
@@ -209,9 +212,9 @@ def test_ssh_raises_exception(mock_run, returncode, error):
 
     mock_run.assert_called_once_with(
         ["ssh", "-tKq", "-o", "BatchMode=yes", "my-host.example.com", f"exit {returncode}"],
-        stdout=ANY,
-        stderr=ANY,
-        stdin=ANY,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
+        stdin=sys.stdin,
     )
 
 
@@ -227,21 +230,12 @@ def test_ssh_returns_error_code_if_configured(mock_run, returncode):
     assert code == returncode
     mock_run.assert_called_once_with(
         ["ssh", "-tKq", "-o", "BatchMode=yes", "my-host.example.com", f"exit {returncode}"],
-        stdout=ANY,
-        stderr=ANY,
-        stdin=ANY,
+        stdout=sys.stdout,
+        stderr=sys.stderr,
+        stdin=sys.stdin,
     )
 
 
-@patch("remote.util.subprocess.run")
-def test_ssh_no_execute_on_dry_run(mock_run):
-    ssh = Ssh("my-host.example.com")
-    code = ssh.execute("exit 1", dry_run=True)
-
-    assert code == 0
-    mock_run.assert_not_called()
-
-
 @pytest.mark.parametrize(
     "command, expected",
     [
@@ -262,9 +256,8 @@ def test_prepare_shell_command(command, expected):
 @pytest.mark.parametrize(
     "port_value, expected_value, exception_raised",
     [
-        (None, None, None),
-        ("5000", (5000, 5000), None),
-        ("5000:5200", (5000, 5200), None),
+        ("5000", ForwardingOptions(5000, 5000), None),
+        ("5000:5200", ForwardingOptions(5000, 5200), None),
         ("bar:foo", None, InvalidInputError),
         ("2.5:100", None, InvalidInputError),
         ("2.6:32:25", None, InvalidInputError),
@@ -274,7 +267,7 @@ def test_prepare_shell_command(command, expected):
 def test_parse_ports(port_value, expected_value, exception_raised):
     if exception_raised:
         with raises(exception_raised):
-            parse_ports(port_value)
+            ForwardingOptions.from_string(port_value)
     else:
-        ports = parse_ports(port_value)
+        ports = ForwardingOptions.from_string(port_value)
         assert expected_value == ports
