@@ -1,3 +1,5 @@
+import sys
+
 from pathlib import Path
 from unittest.mock import ANY, MagicMock, call, patch
 
@@ -5,17 +7,10 @@ import pytest
 
 from remote.configuration import RemoteConfig
 from remote.exceptions import InvalidRemoteHostLabel
+from remote.util import CommunicationOptions, ForwardingOptions
 from remote.workspace import SyncedWorkspace
 
 
-@pytest.fixture
-def workspace(workspace_config):
-    workspace_config.ignores.pull.append("build")
-    working_dir = workspace_config.root / "foo" / "bar"
-    working_dir.mkdir(parents=True)
-    return SyncedWorkspace.from_config(workspace_config, working_dir)
-
-
 def test_create_workspace(workspace_config):
     working_dir = workspace_config.root / "foo" / "bar"
     workspace = SyncedWorkspace.from_config(workspace_config, working_dir)
@@ -82,9 +77,9 @@ def test_clear_remote_workspace(mock_run, workspace):
     # clear should always delete remote root regardless of what the workign dir is
     mock_run.assert_called_once_with(
         ["ssh", "-tKq", "-o", "BatchMode=yes", workspace.remote.host, f"rm -rf {workspace.remote.directory}"],
-        stderr=ANY,
-        stdin=ANY,
-        stdout=ANY,
+        stderr=sys.stderr,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
     )
 
 
@@ -104,9 +99,33 @@ def test_push(mock_run, workspace):
             "--delete",
             "--rsync-path",
             "mkdir -p remote/dir && rsync",
+            "--include-from",
+            ANY,
             f"{workspace.local_root}/",
             f"{workspace.remote.host}:{workspace.remote.directory}",
         ],
+        stderr=sys.stderr,
+        stdout=sys.stdout,
+    )
+
+
+@patch("remote.util.subprocess.run")
+def test_push_with_subdir(mock_run, workspace):
+    mock_run.return_value = MagicMock(returncode=0)
+
+    workspace.push(subpath=Path("some-path"))
+    mock_run.assert_called_once_with(
+        [
+            "rsync",
+            "-arlpmchz",
+            "--copy-unsafe-links",
+            "-e",
+            "ssh -Kq -o BatchMode=yes",
+            "--force",
+            "--delete",
+            f"{workspace.local_root}/foo/bar/some-path",
+            f"{workspace.remote.host}:{workspace.remote.directory}/foo/bar/",
+        ],
         stderr=ANY,
         stdout=ANY,
     )
@@ -130,8 +149,8 @@ def test_pull(mock_run, workspace):
             f"{workspace.remote.host}:{workspace.remote.directory}/",
             f"{workspace.local_root}",
         ],
-        stderr=ANY,
-        stdout=ANY,
+        stderr=sys.stderr,
+        stdout=sys.stdout,
     )
 
 
@@ -139,6 +158,28 @@ def test_pull(mock_run, workspace):
 def test_pull_with_subdir(mock_run, workspace):
     mock_run.return_value = MagicMock(returncode=0)
 
+    workspace.pull(subpath=Path("some-path"))
+    mock_run.assert_called_once_with(
+        [
+            "rsync",
+            "-arlpmchz",
+            "--copy-unsafe-links",
+            "-e",
+            "ssh -Kq -o BatchMode=yes",
+            "--force",
+            f"{workspace.remote.host}:{workspace.remote.directory}/foo/bar/some-path",
+            f"{workspace.local_root}/foo/bar/",
+        ],
+        stderr=sys.stderr,
+        stdout=sys.stdout,
+    )
+
+
+@patch("remote.util.subprocess.run")
+def test_pull_with_subdir_exec_from_root(mock_run, workspace):
+    workspace.remote_working_dir = workspace.remote.directory
+    mock_run.return_value = MagicMock(returncode=0)
+
     workspace.pull(subpath=Path("some-path"))
     mock_run.assert_called_once_with(
         [
@@ -151,8 +192,8 @@ def test_pull_with_subdir(mock_run, workspace):
             f"{workspace.remote.host}:{workspace.remote.directory}/some-path",
             f"{workspace.local_root}/",
         ],
-        stderr=ANY,
-        stdout=ANY,
+        stderr=sys.stderr,
+        stdout=sys.stdout,
     )
 
 
@@ -169,25 +210,72 @@ def test_execute(mock_run, workspace):
             "BatchMode=yes",
             workspace.remote.host,
             """\
-if [ -f remote/dir/.remoteenv ]; then
-  source remote/dir/.remoteenv 2>/dev/null 1>/dev/null
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
 fi
-cd remote/dir/foo/bar
+cd foo/bar
 echo 'Hello World!'
 """,
         ],
-        stderr=ANY,
-        stdin=ANY,
-        stdout=ANY,
+        stderr=sys.stderr,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
     )
     assert code == 0
 
 
+@patch("remote.util.subprocess.run")
+def test_execute_with_dry_run(mock_run, workspace):
+    mock_run.return_value = MagicMock(returncode=0)
+
+    code = workspace.execute(["echo", "Hello World!"], dry_run=True)
+    mock_run.assert_called_once_with(
+        ["ssh", "-tKq", "-o", "BatchMode=yes", workspace.remote.host, "echo echo 'Hello World!'"],
+        stderr=sys.stderr,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+    )
+    assert code == 0
+
+
+@patch("remote.util.subprocess.run")
+def test_execute_with_communication_override(mock_run, workspace, tmp_path):
+    mock_run.return_value = MagicMock(returncode=0)
+    logs_dir = tmp_path / "logs"
+    logs_dir.mkdir(parents=True)
+
+    with (logs_dir / "output.log").open("w") as output_file:
+        workspace.communication = CommunicationOptions(stderr=output_file, stdout=output_file, stdin=None)
+        code = workspace.execute(["echo", "Hello World!"])
+        mock_run.assert_called_once_with(
+            [
+                "ssh",
+                "-tKq",
+                "-o",
+                "BatchMode=yes",
+                workspace.remote.host,
+                """\
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd foo/bar
+echo 'Hello World!'
+""",
+            ],
+            stderr=output_file,
+            stdin=None,
+            stdout=output_file,
+        )
+        assert code == 0
+
+
 @patch("remote.util.subprocess.run")
 def test_execute_with_port_forwarding(mock_run, workspace):
     mock_run.return_value = MagicMock(returncode=0)
 
-    code = workspace.execute(["echo", "Hello World!"], ports=(5005, 5000),)
+    code = workspace.execute(["echo", "Hello World!"], ports=ForwardingOptions(5005, 5000),)
     mock_run.assert_called_once_with(
         [
             "ssh",
@@ -198,16 +286,80 @@ def test_execute_with_port_forwarding(mock_run, workspace):
             "5000:localhost:5005",
             workspace.remote.host,
             """\
-if [ -f remote/dir/.remoteenv ]; then
-  source remote/dir/.remoteenv 2>/dev/null 1>/dev/null
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
 fi
-cd remote/dir/foo/bar
+cd foo/bar
 echo 'Hello World!'
 """,
         ],
-        stderr=ANY,
-        stdin=ANY,
-        stdout=ANY,
+        stderr=sys.stderr,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+    )
+    assert code == 0
+
+
+@patch("remote.util.subprocess.run")
+def test_execute_with_custom_port(mock_run, workspace):
+    mock_run.return_value = MagicMock(returncode=0)
+
+    workspace.remote.port = 4321
+    code = workspace.execute(["echo", "Hello World!"], ports=ForwardingOptions(5005, 5000, local_interface="0.0.0.0"),)
+    mock_run.assert_called_once_with(
+        [
+            "ssh",
+            "-tKq",
+            "-o",
+            "BatchMode=yes",
+            "-p",
+            "4321",
+            "-L",
+            "0.0.0.0:5000:localhost:5005",
+            workspace.remote.host,
+            """\
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd foo/bar
+echo 'Hello World!'
+""",
+        ],
+        stderr=sys.stderr,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
+    )
+    assert code == 0
+
+
+@patch("remote.util.subprocess.run")
+def test_execute_with_custom_env(mock_run, workspace):
+    mock_run.return_value = MagicMock(returncode=0)
+
+    code = workspace.execute(["echo", "Hello World!"], env={"TEST_VAR": "test", "OTHER_VAR": "meow"})
+    mock_run.assert_called_once_with(
+        [
+            "ssh",
+            "-tKq",
+            "-o",
+            "BatchMode=yes",
+            workspace.remote.host,
+            """\
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd foo/bar
+export OTHER_VAR=meow
+export TEST_VAR=test
+echo 'Hello World!'
+""",
+        ],
+        stderr=sys.stderr,
+        stdin=sys.stdin,
+        stdout=sys.stdout,
     )
     assert code == 0
 
@@ -230,11 +382,13 @@ def test_execute_and_sync(mock_run, workspace):
                     "--delete",
                     "--rsync-path",
                     "mkdir -p remote/dir && rsync",
+                    "--include-from",
+                    ANY,
                     f"{workspace.local_root}/",
                     f"{workspace.remote.host}:{workspace.remote.directory}",
                 ],
-                stderr=ANY,
-                stdout=ANY,
+                stderr=sys.stderr,
+                stdout=sys.stdout,
             ),
             call(
                 [
@@ -244,16 +398,17 @@ def test_execute_and_sync(mock_run, workspace):
                     "BatchMode=yes",
                     workspace.remote.host,
                     """\
-if [ -f remote/dir/.remoteenv ]; then
-  source remote/dir/.remoteenv 2>/dev/null 1>/dev/null
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
 fi
-cd remote/dir/foo/bar
+cd foo/bar
 echo 'Hello World!'
 """,
                 ],
-                stderr=ANY,
-                stdin=ANY,
-                stdout=ANY,
+                stderr=sys.stderr,
+                stdin=sys.stdin,
+                stdout=sys.stdout,
             ),
             call(
                 [
@@ -268,8 +423,8 @@ echo 'Hello World!'
                     f"{workspace.remote.host}:{workspace.remote.directory}/",
                     f"{workspace.local_root}",
                 ],
-                stderr=ANY,
-                stdout=ANY,
+                stderr=sys.stderr,
+                stdout=sys.stdout,
             ),
         ]
     )
@@ -280,7 +435,7 @@ echo 'Hello World!'
 def test_execute_and_sync_with_port_forwarding(mock_run, workspace):
     mock_run.side_effect = [MagicMock(returncode=0), MagicMock(returncode=10), MagicMock(returncode=0)]
 
-    code = workspace.execute_in_synced_env(["echo", "Hello World!"], ports=(5005, 5000),)
+    code = workspace.execute_in_synced_env(["echo", "Hello World!"], ports=ForwardingOptions(5005, 5000),)
     mock_run.assert_has_calls(
         [
             call(
@@ -294,11 +449,13 @@ def test_execute_and_sync_with_port_forwarding(mock_run, workspace):
                     "--delete",
                     "--rsync-path",
                     "mkdir -p remote/dir && rsync",
+                    "--include-from",
+                    ANY,
                     f"{workspace.local_root}/",
                     f"{workspace.remote.host}:{workspace.remote.directory}",
                 ],
-                stderr=ANY,
-                stdout=ANY,
+                stderr=sys.stderr,
+                stdout=sys.stdout,
             ),
             call(
                 [
@@ -310,16 +467,17 @@ def test_execute_and_sync_with_port_forwarding(mock_run, workspace):
                     "5000:localhost:5005",
                     workspace.remote.host,
                     """\
-if [ -f remote/dir/.remoteenv ]; then
-  source remote/dir/.remoteenv 2>/dev/null 1>/dev/null
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
 fi
-cd remote/dir/foo/bar
+cd foo/bar
 echo 'Hello World!'
 """,
                 ],
-                stderr=ANY,
-                stdin=ANY,
-                stdout=ANY,
+                stderr=sys.stderr,
+                stdin=sys.stdin,
+                stdout=sys.stdout,
             ),
             call(
                 [
@@ -334,9 +492,80 @@ echo 'Hello World!'
                     f"{workspace.remote.host}:{workspace.remote.directory}/",
                     f"{workspace.local_root}",
                 ],
-                stderr=ANY,
-                stdout=ANY,
+                stderr=sys.stderr,
+                stdout=sys.stdout,
             ),
         ]
     )
     assert code == 10
+
+
+@patch("remote.util.subprocess.run")
+def test_execute_and_sync_with_communication_override(mock_run, workspace, tmp_path):
+    mock_run.side_effect = [MagicMock(returncode=0), MagicMock(returncode=10), MagicMock(returncode=0)]
+    logs_dir = tmp_path / "logs"
+    logs_dir.mkdir(parents=True)
+
+    with (logs_dir / "output.log").open("w") as output_file:
+        workspace.communication = CommunicationOptions(stderr=output_file, stdout=output_file, stdin=None)
+        code = workspace.execute_in_synced_env(["echo", "Hello World!"])
+        mock_run.assert_has_calls(
+            [
+                call(
+                    [
+                        "rsync",
+                        "-arlpmchz",
+                        "--copy-unsafe-links",
+                        "-e",
+                        "ssh -Kq -o BatchMode=yes",
+                        "--force",
+                        "--delete",
+                        "--rsync-path",
+                        "mkdir -p remote/dir && rsync",
+                        "--include-from",
+                        ANY,
+                        f"{workspace.local_root}/",
+                        f"{workspace.remote.host}:{workspace.remote.directory}",
+                    ],
+                    stderr=output_file,
+                    stdout=output_file,
+                ),
+                call(
+                    [
+                        "ssh",
+                        "-tKq",
+                        "-o",
+                        "BatchMode=yes",
+                        workspace.remote.host,
+                        """\
+cd remote/dir
+if [ -f .remoteenv ]; then
+  source .remoteenv
+fi
+cd foo/bar
+echo 'Hello World!'
+""",
+                    ],
+                    stderr=output_file,
+                    stdin=None,
+                    stdout=output_file,
+                ),
+                call(
+                    [
+                        "rsync",
+                        "-arlpmchz",
+                        "--copy-unsafe-links",
+                        "-e",
+                        "ssh -Kq -o BatchMode=yes",
+                        "--force",
+                        "--exclude-from",
+                        ANY,
+                        f"{workspace.remote.host}:{workspace.remote.directory}/",
+                        f"{workspace.local_root}",
+                    ],
+                    stderr=output_file,
+                    stdout=output_file,
+                ),
+            ]
+        )
+        assert code == 10
